use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.007005
use Test::Spelling 0.12;
use Pod::Wordlist;


add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib ) );
__DATA__
API
Abstract
Builder
CPAN
ContentType
Courriel
DROLSKY
DROLSKY's
DateTime
Dave
Disposition
Email
Gregory
Header
HeaderAttribute
HeaderWithAttributes
Headers
Helpers
Internal
Multipart
Oschwald
Part
PayPal
Postel
Ricardo
Role
Rolsky
Rolsky's
Signes
Single
Streams
Types
UTC
UTF
Zbigniew
ascii
attribute's
autarch
cid
distro
drolsky
filename
goschwald
html
inline
lib
lookup
multipart
rjbs
utf
zh
zzbbyy
Łukasiak
